"""Test the TamperEvidentFile class."""

import os
import tempfile
import unittest

from edx_lint.tamper_evident import TamperEvidentFile


class TamperEvidentFileTest(unittest.TestCase):
    """Tests for TamperEvidentFile."""

    def temp_filename(self):
        """
        Make a temporary filename that will be deleted after the test.

        Returns:
            The name of the temp file.

        """
        fdesc, filename = tempfile.mkstemp(suffix=".txt", prefix="tamper_evident_")
        os.close(fdesc)
        self.addCleanup(os.remove, filename)
        return filename

    def write_tamper_evident(self, text, **kwargs):
        """
        Helper to write a tamper-evident temp file.

        Args:
            text (byte string): the content of the file.

            kwargs: any other arguments to `TamperEvidentFile.write`.

        Returns:
            The name of the temp file.

        """
        filename = self.temp_filename()
        TamperEvidentFile(filename).write(text, **kwargs)
        return filename

    def test_writing(self):
        # The contents are written, with a hash.
        # Different contents produce different hashes.
        filename1 = self.write_tamper_evident(b"Hello!")

        with open(filename1, "rb") as f:
            self.assertEqual(f.read(), b"Hello!\n# a8d191538209e335154750d2df575b9ddfb16fc7\n")

        filename2 = self.write_tamper_evident(b"Hello?")

        with open(filename2, "rb") as f:
            self.assertEqual(f.read(), b"Hello?\n# 4820175d44ef1a2c92e52bd1b3b7f05020d66e1c\n")

    def test_hashline_formatting(self):
        filename1 = self.write_tamper_evident(b"Hello!", hashline="XXX {} YYY")

        with open(filename1, "rb") as f:
            self.assertEqual(f.read(), b"Hello!\nXXX a8d191538209e335154750d2df575b9ddfb16fc7 YYY\n")

    def test_validating_a_good_file(self):
        filename = self.write_tamper_evident(b"Am I OK?")
        tef = TamperEvidentFile(filename)
        self.assertTrue(tef.validate())

    def test_appending_is_detected(self):
        filename = self.write_tamper_evident(b"Am I OK?")

        with open(filename, "ab") as f:
            f.write(b"tamper\n")

        tef = TamperEvidentFile(filename)
        self.assertFalse(tef.validate())

    def test_editing_is_detected(self):
        filename = self.write_tamper_evident(b"Line 1\nLine 2\nLine 3\n")
        with open(filename, "rb") as f:
            text = f.read()
        with open(filename, "wb") as f:
            f.write(b"X")
            f.write(text[1:])
        tef = TamperEvidentFile(filename)
        self.assertFalse(tef.validate())

    def test_oneline_file_is_detected(self):
        filename = self.write_tamper_evident(b"Am I OK?")

        with open(filename, "wb") as f:
            f.write(b"tamper")

        tef = TamperEvidentFile(filename)
        self.assertFalse(tef.validate())



"""Checker for incorrect string translation functions."""

import astroid
from pylint.checkers import BaseChecker, utils

from .common import BASE_ID, check_visitors


def register_checkers(linter):
    """Register checkers."""
    linter.register_checker(TranslationStringConstantsChecker(linter))


@check_visitors
class TranslationStringConstantsChecker(BaseChecker):
    """
    Checks for i18n translation functions (_, ugettext, ungettext, and many
    others) being called on something that isn't a string literal.

    Bad:
        _("hello {}".format(name))
        ugettext("Hello " + name)
        ugettext(value_from_database)

    OK:
        _("hello {}").format(name)

    The message id is `translation-of-non-string`.

    """

    name = "translation-string-checker"

    TRANSLATION_FUNCTIONS = {
        "_",
        "gettext",
        "ngettext",
        "ngettext_lazy",
        "npgettext",
        "npgettext_lazy",
        "pgettext",
        "pgettext_lazy",
        "ugettext",
        "ugettext_lazy",
        "ugettext_noop",
        "ungettext",
        "ungettext_lazy",
    }

    MESSAGE_ID = "translation-of-non-string"
    msgs = {
        ("E%d10" % BASE_ID): (
            "i18n function %s() must be called with a literal string",
            MESSAGE_ID,
            "i18n functions must be called with a literal string",
        )
    }

    @utils.only_required_for_messages(MESSAGE_ID)
    def visit_call(self, node):
        """Called for every function call in the source code."""
        if not isinstance(node.func, astroid.Name):
            # It isn't a simple name, can't deduce what function it is.
            return

        if node.func.name not in self.TRANSLATION_FUNCTIONS:
            # Not a function we care about.
            return

        first = node.args[0]
        if isinstance(first, astroid.Const):
            if isinstance(first.value, str):
                # The first argument is a constant string! All is well!
                return

        # Bad!
        self.add_message(self.MESSAGE_ID, args=node.func.name, node=node)



"""Tamper-evident text files."""

import hashlib
import re


class TamperEvidentFile:
    """Tamper-evident files.

    Write text to a file with `.write()`.  Later, you can validate that it
    hasn't changed with `.validate()`.

    """

    def __init__(self, filename):
        self.filename = filename

    def write(self, text, hashline="# {}"):
        """
        Write `text` to the file.

        Writes the text to the file, with a final line checksumming the
        contents.  The entire file must be written with one `.write()` call.

        The last line is written with the `hashline` format string, which can
        be changed to accommodate different file syntaxes.

        Both arguments are UTF8 byte strings.

        Arguments:
            text (UTF8 byte string): the contents of the file to write.

            hashline (string): the format of the last line to append
                to the file, with "{}" replaced with the hash.

        """
        if not text.endswith(b"\n"):
            text += b"\n"

        actual_hash = hashlib.sha1(text).hexdigest()

        with open(self.filename, "wb") as f:
            f.write(text)
            f.write(hashline.format(actual_hash).encode("utf8"))
            f.write(b"\n")

    def validate(self):
        """
        Check if the file still has its original contents.

        Returns True if the file is unchanged, False if it has been tampered
        with.
        """

        with open(self.filename, "rb") as f:
            text = f.read()

        start_last_line = text.rfind(b"\n", 0, -1)
        if start_last_line == -1:
            return False

        original_text = text[: start_last_line + 1]
        last_line = text[start_last_line + 1 :]

        expected_hash = hashlib.sha1(original_text).hexdigest().encode("utf8")
        match = re.search(b"[0-9a-f]{40}", last_line)
        if not match:
            return False
        actual_hash = match.group(0)
        return actual_hash == expected_hash



"""The edx_lint write command."""

from edx_lint.write import write_file

def write_main(argv):
    """
    write FILENAME
        Write a local copy of FILENAME using FILENAME_tweaks for local tweaks.
    """
    if len(argv) != 1:
        print("Please provide the name of a file to write.")
        return 1

    filename = argv[0]
    ret = write_file(filename, output_fn=print)
    return ret



"""
Test ignore-long-lines regex
"""
# pylint: disable=toggle-no-name,toggle-non-boolean-default-value,annotation-missing-token
import os

from .pylint_test import run_pylint


def test_ignore_long_lines():
    source = """\
        # Bad
        a = 1  # hello world hello world hello world hello world hello world hello world hello world hello world hello world hello world
        # ..pii: hello world hello world hello world hello world hello world hello world hello world hello world hello world hello world
        '''
        < https://www.test.com/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world>
        hello world hello world hello world hello world hello world hello world hello world hello world hello world hello world hello world
        '''

        # Good
        '''
        .. pii: hello world hello world hello world hello world hello world hello world hello world hello world hello world hello world
        '''
        # .. pii: hello world hello world hello world hello world hello world hello world hello world hello world hello world hello world
        '''
        <https://www.test.com/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world>
        https://www.test.com/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world
        '''
        # <https://www.test.com/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world>
        # https://www.test.com/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world
        # .. toggle_description: this is a good toggle, it goes back and forth, toggling a lot, which is what we expect toggles to do.
    """
    msg_ids = "line-too-long"
    rcfile = os.path.join(os.path.dirname(__file__), "..", "..", "pylintrc")
    messages = run_pylint(source, msg_ids, f"--rcfile={rcfile}")

    # normalize the messages by stripping off the 3rd segment.
    messages = {":".join(message.split(":")[:2]) for message in messages}

    expected = {"2:line-too-long", "3:line-too-long", "5:line-too-long", "6:line-too-long"}
    assert expected == messages



"""Check that getattr and setattr aren't being used with literal attribute names."""

import re

import astroid
from pylint.checkers import BaseChecker, utils

from .common import BASE_ID, check_visitors


def register_checkers(linter):
    """Register checkers."""
    linter.register_checker(GetSetAttrLiteralChecker(linter))


@check_visitors
class GetSetAttrLiteralChecker(BaseChecker):
    """
    Checks for string literals used as attribute names with getattr and
    friends. `getattr`, `setattr` and `delattr` should be used to get, set and delete attributes of object
    when names of those attributes are not known beforehand and needs to be determined at runtime.
    Instead of using string literals in these methods standard attribute access/assignment
    operations should be used. e.g.
    `x = obj.attr_name`
    `obj.attr_name = value`
    `del obj.attr_name`
    However, in case of getattr literal attribute name is allowed when a default value is set.

    Bad:
        x = getattr(obj, "attr_name")
        setattr(obj, "attr_name", value)
        delattr(obj, "attr_name")

    OK:
        x = getattr(obj, "attr_name", default_value)

    The message id is literal-used-as-attribute.

    """
    name = "getattr-literal-checker"

    MESSAGE_ID = "literal-used-as-attribute"
    msgs = {
        ("C%d30" % BASE_ID): (
            "%s using a literal attribute name",
            MESSAGE_ID,
            "getattr or setattr using with a literal attribute name",
        )
    }

    @utils.only_required_for_messages(MESSAGE_ID)
    def visit_call(self, node):
        """Called for every function call in the source code."""
        if not isinstance(node.func, astroid.Name):
            # It isn't a simple name, can't deduce what function it is.
            return

        if node.func.name == "getattr":
            if len(node.args) != 2:
                # We only attend to 2-argument getattr()
                return
        elif node.func.name in ["setattr", "delattr"]:
            pass
        else:
            # Not a function we care about.
            return

        second = node.args[1]
        if isinstance(second, astroid.Const):
            if isinstance(second.value, str):
                # The second argument is a constant string! Might be bad!
                # Check the string value: if it's an identifier, then no need
                # for getattr.
                if re.search(r"^[a-zA-Z_][a-zA-Z0-9_]*$", second.value):
                    self.add_message(self.MESSAGE_ID, args=node.func.name, node=node)

        # All is well.



"""Test unittest_assert_check.py"""

import pytest

from .pylint_test import run_pylint


def test_bad_asserts():
    source = """\
        import unittest

        class TestUnittestAssertions(unittest.TestCase):
            def test_wrong_usage(self):
                # This is the usage of unittest assert functions which shouldn't be used.
                self.assertEqual('foo'.upper(), 'FOO')

                true = True
                self.assertTrue(true)
                self.assertFalse(not true)

                self.assertIn("a", "lala")
                self.assertNotIn("b", "lala")

                self.assertGreater(1, 0)
                self.assertLess(1, 2)
        """
    messages = run_pylint(source, "avoid-unittest-asserts", '--load-plugins=edx_lint.pylint.unittest_assert')
    assert messages


def test_good_asserts():
    source = """\
        import unittest

        class TestPytestAssertions(unittest.TestCase):
            def test_right_usage(self):
                # This is the usage of pytest assert functions.
                assert 'foo'.upper() == 'FOO'

                true = True
                assert true
                assert not true

                assert "a" in "lala"
                assert "b" not in "lala"

                assert 1 > 0
                assert 1 < 2
        """
    messages = run_pylint(source, "avoid-unittest-asserts", '--load-plugins=edx_lint.pylint.unittest_assert')
    assert not messages


@pytest.mark.parametrize(
    "code, error",
    [
        (
            "assertTrue('foo'.upper() == 'FOO')",
            "assertTrue should be replaced with a pytest assertion something like `assert arg1`"
        ),
        (
            "assertFalse(500 == 501)",
            "assertFalse should be replaced with a pytest assertion something like `assert not arg1`"
        ),
        (
            "assertIn('a', 'lala')",
            "assertIn should be replaced with a pytest assertion something like `assert arg1 in arg2`"
        ),
        (
            "assertIsInstance(1, int)",
            "assertIsInstance should be replaced with a pytest assertion something like `assert isinstance(arg1, arg2)`"
        ),
        (
            "assertEqual('lala', 'lala')",
            "assertEqual should be replaced with a pytest assertion something like `assert arg1 == arg2`"
        ),
        (
            "assertAlmostEqual(6.999, 7)",
            "assertAlmostEqual should be replaced with a pytest assertion something like `assert math.isclose(arg1, "
            "arg2)`"
        ),
        (
            "assertIsNone(somevar)",
            "assertIsNone should be replaced with a pytest assertion something like `assert arg1 is None`"
        ),
    ],
)
def test_assert_hints(code, error):
    source = (
        """\
        import unittest

        class TestUnittestAssertions(unittest.TestCase):
            def test_wrong_usage(self):
                self.{}      #=A
        """
    ).format(code)
    messages = run_pylint(source, "avoid-unittest-asserts", '--load-plugins=edx_lint.pylint.unittest_assert')

    expected = {f"A:avoid-unittest-asserts:{error}"}
    assert expected == messages



"""Test yaml_load_check.py"""

from .pylint_test import run_pylint


MSG_IDS = "unsafe-yaml-load"


def test_unsafe_yaml_load_warnings():
    source = """\
        yaml.load('foo.bar')                      #=A
        yaml.load_all('foo.bar')                  #=B

    """
    messages = run_pylint(source, MSG_IDS)

    expected_messages = {
        "A:unsafe-yaml-load:yaml.load() call is unsafe, use yaml.safe_load()",
        "B:unsafe-yaml-load:yaml.load_all() call is unsafe, use yaml.safe_load_all()",
    }
    assert expected_messages == messages


def test_safe_yaml_invocations_are_fine():
    source = """\
        yaml.safe_load('this.is.fine')
        yaml.safe_load_all('this.is.fine')
        myaml.load('this.is.fine.too')
        myaml.load_all('this.is.fine.too')

    """
    messages = run_pylint(source, MSG_IDS)
    assert not messages



"""Checker for simplifiable range() calls."""

import astroid

from pylint.checkers import BaseChecker, utils

from .common import BASE_ID, check_visitors


def register_checkers(linter):
    """Register checkers."""
    linter.register_checker(RangeChecker(linter))


@check_visitors
class RangeChecker(BaseChecker):
    """
    Checks for range() and xrange() used with unneeded arguments.

    Bad:
        range(0, N)
        range(0, N, 1)

    OK:
        range(N)
        range(1, N)
        range(1, N, 2)

    The message id is simplifiable-range.

    """

    name = "range-checker"

    RANGE_FUNCTIONS = ["range", "xrange"]

    MESSAGE_ID = "simplifiable-range"
    msgs = {("C%d20" % BASE_ID): ("%s() call could be %s-argument", MESSAGE_ID, "range() call could be simplified")}

    @utils.only_required_for_messages(MESSAGE_ID)
    def visit_call(self, node):
        """Called for every function call in the source code."""
        if not isinstance(node.func, astroid.Name):
            # It isn't a simple name, can't deduce what function it is.
            return

        if node.func.name not in self.RANGE_FUNCTIONS:
            # Not a function we care about.
            return

        first = node.args[0]
        if not isinstance(first, astroid.Const):
            # Computed first argument, can't tell what it is.
            return

        if not isinstance(first.value, int):
            # First argument is not an int, that's fine.
            return

        # If there are three args and the third is 1, that's bad.
        three1 = False
        if len(node.args) == 3:
            third = node.args[2]
            if isinstance(third, astroid.Const):
                if isinstance(third.value, int) and third.value == 1:
                    three1 = True

        if first.value == 0:
            # The first argument is 0, suspicious.
            if len(node.args) == 2:
                # range(0, n): bad.
                self.add_message(self.MESSAGE_ID, args=(node.func.name, "single"), node=node)
            elif three1:
                # range(0, n, 1): bad.
                self.add_message(self.MESSAGE_ID, args=(node.func.name, "single"), node=node)
        elif three1:
            # range(n, m, 1): bad.
            self.add_message(self.MESSAGE_ID, args=(node.func.name, "two"), node=node)



"""Pylint plugin: check that tests have used super() properly."""

import astroid
from pylint.checkers import BaseChecker, utils
try:
    from pylint.checkers.classes import _ancestors_to_call
except ImportError:  # Backward compatibility with pylint<2.13
    from pylint.checkers.classes.class_checker import _ancestors_to_call

from .common import BASE_ID, check_visitors, usable_class_name


def register_checkers(linter):
    """Register checkers."""
    linter.register_checker(UnitTestSetupSuperChecker(linter))


@check_visitors
class UnitTestSetupSuperChecker(BaseChecker):
    """
    Checks that unittest methods have used super() properly.

    It examines `setUp`, `tearDown`, `setUpClass`, and `tearDownClass` for
    correct use of super.  If there is no super, it issues a
    `super-method-not-called` error.  If super is used, but with the wrong
    class name, it issues a `non-parent-method-called` error.

    """

    name = "unit-test-super-checker"

    NOT_CALLED_MESSAGE_ID = "super-method-not-called"
    NON_PARENT_MESSAGE_ID = "non-parent-method-called"

    METHOD_NAMES = ["setUp", "tearDown", "setUpClass", "tearDownClass", "setUpTestData"]

    msgs = {
        ("E%d01" % BASE_ID): (
            "super(...).%s() not called (%s)",
            NOT_CALLED_MESSAGE_ID,
            "setUp() must call super(...).setUp()",
        ),
        ("E%d02" % BASE_ID): (
            "%s() was called from a non-parent class (%s)",
            NON_PARENT_MESSAGE_ID,
            "setUp() should only be called for parent classes",
        ),
    }

    @utils.only_required_for_messages(NOT_CALLED_MESSAGE_ID, NON_PARENT_MESSAGE_ID)
    def visit_functiondef(self, node):
        """Called for every function definition in the source code."""
        # ignore actual functions
        if not node.is_method():
            return

        method_name = node.name
        if method_name not in self.METHOD_NAMES:
            return

        klass_node = node.parent.frame()
        to_call = _ancestors_to_call(klass_node, method_name)

        not_called_yet = dict(to_call)
        for stmt in node.nodes_of_class(astroid.Call):
            expr = stmt.func
            if not isinstance(expr, astroid.Attribute):
                continue
            if expr.attrname != method_name:
                continue

            # Skip the test if using super
            if (
                isinstance(expr.expr, astroid.Call)
                and isinstance(expr.expr.func, astroid.Name)
                and expr.expr.func.name == "super"
            ):
                return

            try:
                klass = next(expr.expr.infer())
                if klass is astroid.Uninferable:
                    continue

                # The infered klass can be super(), which was
                # assigned to a variable and the `__init__` was called later.
                #
                # base = super()
                # base.__init__(...)

                # pylint: disable=protected-access
                if (
                    isinstance(klass, astroid.Instance)
                    and isinstance(klass._proxied, astroid.ClassDef)
                    and utils.is_builtin_object(klass._proxied)
                    and klass._proxied.name == "super"
                ):
                    return
                if isinstance(klass, astroid.objects.Super):
                    return
                try:
                    del not_called_yet[klass]
                except KeyError:
                    if klass not in to_call:
                        self.add_message(
                            self.NON_PARENT_MESSAGE_ID, node=expr, args=(method_name, usable_class_name(klass))
                        )
            except astroid.InferenceError:
                continue

        for klass, method in not_called_yet.items():
            if klass.name == "object" or method.parent.name == "object":
                continue
            self.add_message(self.NOT_CALLED_MESSAGE_ID, args=(method_name, usable_class_name(klass)), node=node)



"""Test annotations_check.py"""
# pylint: disable=toggle-non-boolean-default-value,toggle-empty-description,toggle-no-name,annotation-missing-token

from .pylint_test import run_pylint


def test_waffle_missing_toggle_annotation_check():
    source = """\
        DisablePragmaWaffleFlag(NAMESPACE, 'disable_pragma_for_annotation')  #pylint: disable=feature-toggle-needs-doc

        # .. toggle_name: annotated_flag
        WaffleFlag(NAMESPACE, 'annotated_flag')

        # .. toggle_name: course_waffle_annotated_flag
        CourseWaffleFlag(NAMESPACE, 'course_waffle_annotated_flag')

        NotAFlag(NAMESPACE, NOT_A_WAFFLE_FLAG)

        # .. wrong_annotation
        WaffleFlag(NAMESPACE, 'flag_with_bad_annotation') #=A

        WaffleFlag(NAMESPACE, FLAG_WITHOUT_ANNOTATION) #=B

        DerivedWaffleFlag(NAMESPACE, DERIVED_FLAG_WITHOUT_ANNOTATION) #=C

        WaffleSwitch(NAMESPACE, SWITCH_WITHOUT_ANNOTATION) #=D

        CourseWaffleFlag(NAMESPACE, COURSE_WAFFLE_FLAG_WITHOUT_ANNOTATION) #=E

        MissingCourseWithKwarg = CourseWaffleFlag( #=F
            waffle_namespace=waffle_flags(),
            flag_name=u'missing_course_with_kwarg',
        )
        """

    msg_ids = "feature-toggle-needs-doc"
    messages = run_pylint(source, msg_ids)
    expected = {
        "A:feature-toggle-needs-doc:feature toggle ('flag_with_bad_annotation') is missing annotation",
        "B:feature-toggle-needs-doc:feature toggle (FLAG_WITHOUT_ANNOTATION) is missing annotation",
        "C:feature-toggle-needs-doc:feature toggle (DERIVED_FLAG_WITHOUT_ANNOTATION) is missing annotation",
        "D:feature-toggle-needs-doc:feature toggle (SWITCH_WITHOUT_ANNOTATION) is missing annotation",
        "E:feature-toggle-needs-doc:feature toggle (COURSE_WAFFLE_FLAG_WITHOUT_ANNOTATION) is missing annotation",
        "F:feature-toggle-needs-doc:feature toggle (missing_course_with_kwarg) is missing annotation",
    }
    assert expected == messages


def test_config_models_missing_doc():
    source = """\
        from config_models.models import ConfigurationModel
        from django.db import models

        # .. toggle_name: my_toggle_name
        class CorrectlyAnnotatedConfig(ConfigurationModel):
            my_toggle_name = models.BooleanField(default=True)

        class DisabeldNoAnnotationsConfig(ConfigurationModel):  #pylint: disable=feature-toggle-needs-doc
            my_toggle_name = models.BooleanField(default=True)

        # .. wrong_annotation
        class WronglyAnnotatedConfig(ConfigurationModel): #=A
            my_toggle_name = models.BooleanField(default=True)

        class NoAnnotationsConfig(ConfigurationModel): #=B
            my_toggle_name = models.BooleanField(default=True)

        class NotAConfigurationModelClass():
            def __init__(self, value):
                self.value = value
        """

    msg_ids = "feature-toggle-needs-doc"
    messages = run_pylint(source, msg_ids)
    expected = {
        "A:feature-toggle-needs-doc:feature toggle (WronglyAnnotatedConfig) is missing annotation",
        "B:feature-toggle-needs-doc:feature toggle (NoAnnotationsConfig) is missing annotation",
    }
    assert expected == messages


def test_django_feature_flags_missing_doc():
    source = """\
        {
            "key_value": "value"
        }

        COURSE_DICT = {
            'COURSE1',
            'COURSE2',
            'COURSE3'
        }

        FEATURES = {  #=A
            # .. toggle_name: CORRECTLY_ANNOTATED_FLAG
            'CORRECTLY_ANNOTATED_FLAG': True,

            'NO_DOCUMENTATION_FLAG': False,

            # .. wrong_annotation
            'WRONG_DOCUMENTATION_FLAG': False,
        }

        FEATURES = {  #pylint: disable=feature-toggle-needs-doc
            'SECOND_NO_DOCUMENTATION_FLAG': False,
        }
        """

    msg_ids = "feature-toggle-needs-doc"
    messages = run_pylint(source, msg_ids)
    expected = {
        "A:feature-toggle-needs-doc:feature toggle (NO_DOCUMENTATION_FLAG) is missing annotation",
        "A:feature-toggle-needs-doc:feature toggle (WRONG_DOCUMENTATION_FLAG) is missing annotation",
    }
    assert expected == messages


def test_illegal_waffle_usage_check():
    source = """\
        switch_is_active('disable_pragma')  #pylint: disable=illegal-waffle-usage

        switch_is_active('test_switch')  #=A

        switch_is_active(TEST_SWITCH)  #=B

        flag_is_active('test_flag')  #=C

        flag_is_active(TEST_FLAG)  #=D
        """

    msg_ids = "illegal-waffle-usage"
    messages = run_pylint(source, msg_ids)
    expected = {
        "A:illegal-waffle-usage:illegal waffle usage with ('test_switch'): "
        "use utility classes WaffleFlag, WaffleSwitch, CourseWaffleFlag.",
        "B:illegal-waffle-usage:illegal waffle usage with (TEST_SWITCH): "
        "use utility classes WaffleFlag, WaffleSwitch, CourseWaffleFlag.",
        "C:illegal-waffle-usage:illegal waffle usage with ('test_flag'): "
        "use utility classes WaffleFlag, WaffleSwitch, CourseWaffleFlag.",
        "D:illegal-waffle-usage:illegal waffle usage with (TEST_FLAG): "
        "use utility classes WaffleFlag, WaffleSwitch, CourseWaffleFlag.",
    }
    assert expected == messages


def test_code_annotations_checker():
    source = """
    # .. toggle_name: MYTOGGLE
    # .. toggle_name: MYTOGGLE
    """
    messages = run_pylint(source, "annotation-duplicate-token")
    expected = {
        "2:annotation-duplicate-token:found duplicate token '.. toggle_name:'"
    }
    assert expected == messages


def test_temporary_use_case_without_target_removal_date():
    source = """
    # .. toggle_name: MYTOGGLE
    # .. toggle_use_cases: temporary
    """
    messages = run_pylint(source, "toggle-missing-target-removal-date")
    expected = {
        "2:toggle-missing-target-removal-date:temporary feature toggle (MYTOGGLE) has no target removal date"
    }
    assert expected == messages


def test_empty_removal_date_on_permanent_use_case():
    source = """
    # .. toggle_name: MYTOGGLE
    # .. toggle_use_cases: open_edx
    """
    messages = run_pylint(source, "toggle-missing-target-removal-date")
    assert not messages


def test_toggle_with_empty_name():
    source = """
    # .. toggle_name:
    """
    messages = run_pylint(source, "toggle-no-name")
    expected = {
        "2:toggle-no-name:feature toggle has no name"
    }
    assert expected == messages


def test_toggle_with_empty_description():
    source = """
    # .. toggle_name: MYTOGGLE
    # .. toggle_description:
    """
    messages = run_pylint(source, "toggle-empty-description")
    expected = {
        "2:toggle-empty-description:feature toggle (MYTOGGLE) does not have a description"
    }
    assert expected == messages


def test_non_boolean_default_value():
    source = """
    # .. toggle_name: MYTOGGLE
    # .. toggle_default: something
    """
    messages = run_pylint(source, "toggle-non-boolean-default-value")
    expected = {
        "2:toggle-non-boolean-default-value:feature toggle (MYTOGGLE) default value must be boolean ('True' or 'False')"
    }
    assert expected == messages


def test_setting_boolean_default_value():
    source = """
    # .. setting_name: MYSETTING
    # .. setting_default: True
    """
    messages = run_pylint(source, "setting-boolean-default-value")
    expected = {
        "2:setting-boolean-default-value:setting annotation (MYSETTING) cannot have a boolean value"
    }
    assert expected == messages


def test_no_duplicate_annotation_errors():
    source = """
    # .. setting_default: something1
    # .. setting_description: something1
    x = 1

    # .. setting_name: MYTOGGLE2
    # .. setting_default: something2
    # .. setting_description: something2
    x = 2
    """
    messages = run_pylint(source, "annotation-missing-token")
    expected = {
        "2:annotation-missing-token:missing non-optional annotation: '.. setting_name:'"
    }
    assert expected == messages


def test_missing_annotation():
    source = """
    # .. toggle_name: MYTOGGLE1
    waffle1 = WaffleFlag('MYTOGGLE1')
    waffle2 = WaffleFlag('MYTOGGLE2')
    """
    messages = run_pylint(source, "toggle-missing-annotation")
    expected = {
        "4:toggle-missing-annotation:missing feature toggle annotation"
    }
    assert expected == messages


def test_missing_annotation_for_unnamed_toggle():
    source = """
    # annotated waffle flag
    # .. toggle_name: MYTOGGLE1
    waffle1 = CourseWaffleFlag()
    # unannotated waffle flag
    waffle2 = ExperimentWaffleFlag()
    """
    messages = run_pylint(source, "toggle-missing-annotation")
    expected = {
        "6:toggle-missing-annotation:missing feature toggle annotation"
    }
    assert expected == messages


def test_invalid_import_from_django_waffle():
    source = """
    from waffle import waffle_is_active
    import waffle
    """
    messages = run_pylint(source, "invalid-django-waffle-import")
    expected = {
        "2:invalid-django-waffle-import:invalid Django Waffle import",
        "3:invalid-django-waffle-import:invalid Django Waffle import",
    }
    assert expected == messages



"""Test right_assert_check.py"""

import pytest

from .pylint_test import run_pylint


def test_good_asserts():
    source = """\
        import unittest

        class TestStringMethods(unittest.TestCase):
            def test_right_usage(self):
                # This is the right usage of various assert functions.
                self.assertEqual('foo'.upper(), 'FOO')

                true = True
                self.assertTrue(true)
                self.assertFalse(not true)

                self.assertIn("a", "lala")
                self.assertNotIn("b", "lala")

                self.assertGreater(1, 0)
                self.assertLess(1, 2)

            def test_chained_comparisons(self):
                # These uses of assertTrue and assertFalse are fine, because we can't
                # pick apart the chained comparisons.
                my_value = my_other_value = 10
                self.assertTrue(0 < my_value < 1000)
                self.assertFalse(0 < my_value < 5)
                self.assertTrue(my_value == my_other_value == 10)

            def test_other_functions(self):
                foo(bar)
                assertTrue("what is this?")
        """
    messages = run_pylint(source, "wrong-assert-type")
    assert not messages


@pytest.mark.parametrize(
    "code, better",
    [
        ("assertTrue('foo'.upper() == 'FOO')", "assertEqual"),
        ("assertFalse(500 == 501)", "assertNotEqual"),
        ("assertTrue('a' in 'lala')", "assertIn"),
        ("assertFalse('b' not in 'lala')", "assertIn"),
        ("assertTrue(1 > 0)", "assertGreater"),
        ("assertFalse(1 < 2)", "assertGreaterEqual"),
        ("assertTrue(my_zero is 0)", "assertIs"),
        ("assertFalse(my_zero is 1)", "assertIsNot"),
        ("assertTrue(my_zero is not 1)", "assertIsNot"),
        ("assertFalse(my_zero is not 0)", "assertIs"),
        ("assertTrue(my_none is None)", "assertIsNone"),
        ("assertFalse(my_zero is None)", "assertIsNotNone"),
        ("assertTrue(my_zero != None)", "assertIsNotNone"),
    ],
)
def test_wrong_usage(code, better):
    source = (
        """\
        import unittest

        class TestStringMethods(unittest.TestCase):
            def test_wrong_usage(self):
                self.{}      #=A
        """
    ).format(code)
    messages = run_pylint(source, "wrong-assert-type")

    expected = {f"A:wrong-assert-type:{code} should be {better}"}
    assert expected == messages



"""Infrastructure for testing pylint plugins."""

import re
import textwrap
import warnings

from pylint.lint import Run
from pylint.reporters import CollectingReporter


def find_line_markers(source):
    """Find line markers in program source.

    Returns a dict mapping line numbers to the marker on that line.
    """
    markers = {}
    for lineno, line in enumerate(source.splitlines(), start=1):
        m = re.search(r"#=(\w+)", line)
        if m:
            markers[lineno] = m.group(1)
    return markers


def test_find_line_markers():
    markers = find_line_markers(
        """\
        line 1      #=A
        line 2
        line 3      #=Hello
        """
    )
    assert markers == {1: "A", 3: "Hello"}


class SimpleReporter(CollectingReporter):
    """A pylint message reporter that collects the messages in a list."""

    # Pylint does not specify well what a reporter must do.  This works.

    def _display(self, layout):
        pass


def run_pylint(source, msg_ids, *cmd_args):
    """Run pylint on some source, collecting specific messages.

    `source` is the literal text of the program to check. It is
    dedented and written to a temp file for pylint to read.

    `msg_ids` is a comma-separated string of msgids we are interested
    in.  Use "all" to enable all messages.

    `*cmd_args` is the optional list of command line arguments. If calling
    function wants to send some additional command line arguments.

    Returns a set of messages.  Each message is a string, formatted
    as "line:msg-id:message".  "line" will be the line number of the
    message, or if the source line has a comment like "#=Slug", then
    it will be "Slug" instead.  This makes it easier to write, read,
    and maintain the tests.

    """
    with open("source.py", "w") as f:
        f.write(textwrap.dedent(source))

    reporter = SimpleReporter()

    pylint_args = ["source.py", "--disable=all", f"--enable={msg_ids}", "--load-plugins=edx_lint.pylint", *cmd_args]

    # --disable=all & --enable=all are not compatible
    if msg_ids == "all":
        pylint_args.remove("--disable=all")

    kwargs = {"exit": False}

    Run(pylint_args, reporter=reporter, **kwargs)

    markers = find_line_markers(source)
    messages = {"{line}:{m.symbol}:{m.msg}".format(m=m, line=markers.get(m.line, m.line)) for m in reporter.messages}
    return messages


def test_that_we_can_test_pylint():
    # This tests that our pylint-testing function works properly.
    source = """\
        # There's no docstring, but we don't ask for that msgid,
        # so we won't get the warning.

        # Unused imports. We'll get warned about the first one,
        # but the second is disabled.
        import colorsys                 #=A
        import collections              # pylint: disable=unused-import

        # Three warnings on the same line, two different messages.
        # redefined-builtin is checked by an IAstroidChecker.
        # anomalous-backslash-in-string is checked by an ITokenChecker.
        int = float = "\\a\\b\\c"       #=B

        # TODO is checked by an IRawChecker. #=C
        """
    msg_ids = "unused-import,redefined-builtin,anomalous-backslash-in-string,fixme"
    with warnings.catch_warnings():
        # We want pylint to find the bad \c escape, but we don't want Python to warn about it.
        warnings.filterwarnings(action="ignore", category=DeprecationWarning, message="invalid escape")
        messages = run_pylint(source, msg_ids)
    expected = {
        "A:unused-import:Unused import colorsys",
        "B:redefined-builtin:Redefining built-in 'int'",
        "B:redefined-builtin:Redefining built-in 'float'",
        "B:anomalous-backslash-in-string:Anomalous backslash in string: '\\c'. "
        "String constant might be missing an r prefix.",
        "C:fixme:TODO is checked by an IRawChecker. #=C",
    }
    assert expected == messages


def test_invalid_python():
    source = """\
        This isn't even Python, what will pylint do?
        """
    messages = run_pylint(source, "all")
    assert len(messages) == 1
    message = messages.pop()
    assert message == "1:syntax-error:Parsing failed: 'invalid syntax (source, line 1)'"


# I would have tested that the msgids must be valid, but pylint doesn't seem
# to mind being told to enable non-existent msgids.



"""
Pylint plugin: checks that feature toggles are properly annotated.
"""

import os
import re

import pkg_resources

from astroid.nodes.node_classes import Const, Name
from code_annotations import annotation_errors
from code_annotations.base import AnnotationConfig
from code_annotations.find_static import StaticSearch
from pylint.checkers import BaseChecker, utils

from .common import BASE_ID, check_visitors


def register_checkers(linter):
    """
    Register checkers.
    """
    linter.register_checker(FeatureToggleChecker(linter))
    linter.register_checker(CodeAnnotationChecker(linter))
    linter.register_checker(FeatureToggleAnnotationChecker(linter))
    linter.register_checker(SettingAnnotationChecker(linter))


def check_all_messages(msgs):
    """
    Decorator to automatically assign all messages from a class to the list of messages handled by a checker method.

    Inspired by pylint.checkers.util.only_required_for_messages
    """

    def store_messages(func):
        func.checks_msgs = [message[1] for message in msgs]
        return func

    return store_messages


class AnnotationLines:
    """
    AnnotationLines provides utility methods to work with a string in terms of
    lines.  As an example, it can convert a Call node into a list of its contents
    separated by line breaks.
    """

    # Regex searches for annotations like: # .. toggle
    _ANNOTATION_REGEX = re.compile(r"[\s]*#[\s]*\.\.[\s]*(toggle)")

    def __init__(self, module_node):
        """
        Arguments:
            module_node: The visited module node.
        """
        module_as_binary = module_node.stream().read()

        file_encoding = module_node.file_encoding
        if file_encoding is None:
            file_encoding = "UTF-8"

        module_as_string = module_as_binary.decode(file_encoding)
        self._list_of_string_lines = module_as_string.split("\n")

    def is_line_annotated(self, line_number):
        """
        Checks if the provided line number is annotated.
        """
        if line_number < 1 or self._line_count() < line_number:
            return False

        return bool(self._ANNOTATION_REGEX.match(self._get_line_contents(line_number)))

    def _line_count(self):
        """
        Gets the number of lines in the string.
        """
        return len(self._list_of_string_lines)

    def _get_line_contents(self, line_number):
        """
        Gets the line of text designated by the provided line number.
        """
        return self._list_of_string_lines[line_number - 1]


@check_visitors
class FeatureToggleChecker(BaseChecker):
    """
    Checks that feature toggles are properly annotated and best practices
    are followed.
    """

    name = "feature-toggle-checker"

    TOGGLE_NOT_ANNOTATED_MESSAGE_ID = "feature-toggle-needs-doc"
    ILLEGAL_WAFFLE_MESSAGE_ID = "illegal-waffle-usage"

    _CHECK_CAPITAL_REGEX = re.compile(r"[A-Z]")
    _WAFFLE_TOGGLE_CLASSES = ("WaffleFlag", "WaffleSwitch", "CourseWaffleFlag")
    _ILLEGAL_WAFFLE_FUNCTIONS = ["flag_is_active", "switch_is_active"]

    msgs = {
        ("E%d40" % BASE_ID): (
            "feature toggle (%s) is missing annotation",
            TOGGLE_NOT_ANNOTATED_MESSAGE_ID,
            "feature toggle is missing annotation",
        ),
        ("E%d41" % BASE_ID): (
            "illegal waffle usage with (%s): use utility classes {}.".format(
                ", ".join(_WAFFLE_TOGGLE_CLASSES)
            ),
            ILLEGAL_WAFFLE_MESSAGE_ID,
            "illegal waffle usage: use utility classes {}.".format(
                ", ".join(_WAFFLE_TOGGLE_CLASSES)
            ),
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lines = None

    def visit_module(self, node):
        """Parses the module code to provide access to comments."""
        self._lines = AnnotationLines(node)

    def check_waffle_class_annotated(self, node):
        """
        Check Call node for waffle class instantiation with missing annotations.
        """
        if not hasattr(node.func, "name"):
            return

        # Looking for class instantiation, so should start with a capital letter
        starts_with_capital = self._CHECK_CAPITAL_REGEX.match(node.func.name)
        if not starts_with_capital:
            return

        # Search for toggle classes that require an annotation
        if not node.func.name.endswith(self._WAFFLE_TOGGLE_CLASSES):
            return

        if not self._lines.is_line_annotated(node.lineno - 1):
            feature_toggle_name = "UNKNOWN"

            if node.keywords is not None:
                for node_key in node.keywords:
                    if node_key.arg == "flag_name":
                        feature_toggle_name = node_key.value.value

            if feature_toggle_name == "UNKNOWN":
                if len(node.args) >= 2:
                    feature_toggle_name = node.args[1].as_string()

            self.add_message(
                self.TOGGLE_NOT_ANNOTATED_MESSAGE_ID,
                args=(feature_toggle_name,),
                node=node,
            )

    def check_configuration_model_annotated(self, node):
        """
        Checks class definitions to see if they subclass ConfigurationModel.
        If they do, they should be correctly annotated.
        """
        if "ConfigurationModel" not in node.basenames:
            return
        if not self._lines.is_line_annotated(node.lineno - 1):
            config_model_subclass_name = node.name

            self.add_message(
                self.TOGGLE_NOT_ANNOTATED_MESSAGE_ID,
                args=(config_model_subclass_name,),
                node=node,
            )

    def check_django_feature_flag_annotated(self, node):
        """
        Checks dictionary definitions to see if the django feature flags
        dict FEATURES is being set. If it is, entries should be
        correctly annotated.
        """
        try:
            parent_target_name = node.parent.targets[0].name
        except AttributeError:
            return

        if parent_target_name == "FEATURES":
            for key, _ in node.items:
                if not self._lines.is_line_annotated(key.lineno - 1):
                    django_feature_toggle_name = key.value

                    self.add_message(
                        self.TOGGLE_NOT_ANNOTATED_MESSAGE_ID,
                        args=(django_feature_toggle_name,),
                        node=node,
                    )

    def check_illegal_waffle_usage(self, node):
        """
        Check Call node for illegal waffle calls.
        """
        if not hasattr(node.func, "name"):
            return

        if node.func.name in self._ILLEGAL_WAFFLE_FUNCTIONS:
            feature_toggle_name = "UNKNOWN"
            if len(node.args) >= 1:
                feature_toggle_name = node.args[0].as_string()

            self.add_message(
                self.ILLEGAL_WAFFLE_MESSAGE_ID, args=(feature_toggle_name,), node=node
            )

    @utils.only_required_for_messages(TOGGLE_NOT_ANNOTATED_MESSAGE_ID, ILLEGAL_WAFFLE_MESSAGE_ID)
    def visit_call(self, node):
        """
        Performs various checks on Call nodes.
        """
        self.check_waffle_class_annotated(node)
        self.check_illegal_waffle_usage(node)

    @utils.only_required_for_messages(TOGGLE_NOT_ANNOTATED_MESSAGE_ID)
    def visit_classdef(self, node):
        """
        Checks class definitions for potential ConfigurationModel
        implementations.
        """
        self.check_configuration_model_annotated(node)

    @utils.only_required_for_messages(TOGGLE_NOT_ANNOTATED_MESSAGE_ID)
    def visit_dict(self, node):
        """
        Checks Dict nodes in case a Django FEATURES dictionary is being
        initialized.
        """
        self.check_django_feature_flag_annotated(node)


@check_visitors
class AnnotationBaseChecker(BaseChecker):
    """
    Code annotation checkers should almost certainly inherit from this class.

    The CONFIG_FILENAMES class attribute is a list of str filenames located in code_annotations/contrib/config.
    """

    # Override this in child classes
    CONFIG_FILENAMES = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config_search = []
        for config_filename in self.CONFIG_FILENAMES:
            config_path = pkg_resources.resource_filename(
                "code_annotations",
                os.path.join("contrib", "config", config_filename),
            )
            config = AnnotationConfig(config_path, verbosity=-1)
            search = StaticSearch(config)
            self.config_search.append((config, search))
            self.current_module_annotations = []

    def check_module(self, node):
        """
        Perform checks on all annotation groups for this module.
        """
        for config, search in self.config_search:
            # This is a hack to avoid re-creating AnnotationConfig every time
            config.source_path = node.path[0]
            all_results = search.search()

            for _file_name, results in all_results.items():
                for annotations_group in search.iter_groups(results):
                    self.current_module_annotations.append(annotations_group)
                    self.check_annotation_group(search, annotations_group, node)

    def leave_module(self, _node):
        self.current_module_annotations.clear()

    def check_annotation_group(self, search, annotations, node):
        raise NotImplementedError


class CodeAnnotationChecker(AnnotationBaseChecker):
    """
    Run generic code annotation checks.

    This makes use of code-annotations `check_results()` method. Elements from
    `search.annotation_errors` are then parsed and exposed to pylint. Note that modifying the creation order of the
    error types in code-annotations will lead to a modification of the numerical IDs of the errors here, so this should
    be avoided as much as possible.

    When creating a new annotation configuration, its filename should be added to
    CodeAnnotationChecker.CONFIG_FILENAMES (see AnnotationBaseChecker docs).
    """
    CONFIG_FILENAMES = ["feature_toggle_annotations.yaml", "setting_annotations.yaml"]
    name = "code-annotations"
    msgs = {
        ("E%d%d" % (BASE_ID, index + 50)): (
            error_type.message,
            error_type.symbol,
            error_type.description,
        )
        for index, error_type in enumerate(annotation_errors.TYPES)
    }

    @check_all_messages(msgs)
    def visit_module(self, node):
        """
        Run all checks on a single module.
        """
        self.check_module(node)

    def check_annotation_group(self, search, annotations, node):
        search.check_group(annotations)
        for (annotation, AnnotationErrorType, args) in search.annotation_errors:
            self.add_message(
                AnnotationErrorType.symbol,
                args=args,
                node=node,
                line=annotation["line_number"],
            )
        search.annotation_errors.clear()


class FeatureToggleAnnotationChecker(AnnotationBaseChecker):
    """
    Parse feature toggle annotations and ensure best practices are followed.
    """

    CONFIG_FILENAMES = ["feature_toggle_annotations.yaml"]

    name = "toggle-annotations"

    NO_NAME_MESSAGE_ID = "toggle-no-name"
    EMPTY_DESCRIPTION_MESSAGE_ID = "toggle-empty-description"
    MISSING_TARGET_REMOVAL_DATE_MESSAGE_ID = "toggle-missing-target-removal-date"
    NON_BOOLEAN_DEFAULT_VALUE = "toggle-non-boolean-default-value"
    MISSING_ANNOTATION = "toggle-missing-annotation"
    INVALID_DJANGO_WAFFLE_IMPORT = "invalid-django-waffle-import"

    msgs = {
        ("E%d60" % BASE_ID): (
            "feature toggle has no name",
            NO_NAME_MESSAGE_ID,
            "Feature toggle name must be present and be the first annotation",
        ),
        ("E%d61" % BASE_ID): (
            "feature toggle (%s) does not have a description",
            EMPTY_DESCRIPTION_MESSAGE_ID,
            "Feature toggles must include a thorough description",
        ),
        ("E%d62" % BASE_ID): (
            "temporary feature toggle (%s) has no target removal date",
            MISSING_TARGET_REMOVAL_DATE_MESSAGE_ID,
            "Temporary feature toggles must include a target removal date",
        ),
        ("E%d63" % BASE_ID): (
            "feature toggle (%s) default value must be boolean ('True' or 'False')",
            NON_BOOLEAN_DEFAULT_VALUE,
            "Feature toggle default values must be boolean",
        ),
        ("E%d64" % BASE_ID): (
            "missing feature toggle annotation",
            MISSING_ANNOTATION,
            (
                "When a WaffleFlag/Switch object is created, a corresponding annotation must be present above in the"
                " same module and with a matching name",
            )
        ),
        ("E%d65" % BASE_ID): (
            "invalid Django Waffle import",
            INVALID_DJANGO_WAFFLE_IMPORT,
            (
                "Do not directly access Django Waffle objects and methods. Instead, import from"
                " edx_toggles.toggles.",
            )
        ),
    }

    TOGGLE_FUNC_NAMES = [
        "WaffleFlag",
        "NonNamespacedWaffleFlag",
        "WaffleSwitch",
        "NonNamespacedWaffleSwitch",
        "CourseWaffleFlag",
        "ExperimentWaffleFlag",
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current_module_annotated_toggle_names = set()
        self.current_module_annotation_group_line_numbers = []

    @check_all_messages(msgs)
    def visit_module(self, node):
        """
        Run all checks on a single module.
        """
        self.check_module(node)

    def leave_module(self, _node):
        self.current_module_annotated_toggle_names.clear()
        self.current_module_annotation_group_line_numbers.clear()

    def check_annotation_group(self, search, annotations, node):
        """
        Perform checks on a single annotation group.
        """
        if not annotations:
            return

        target_removal_date = None
        temporary_use_case = False
        toggle_name = ""
        toggle_description = ""
        toggle_default = None
        line_number = None
        for annotation in annotations:
            if line_number is None:
                line_number = annotation["line_number"]
                self.current_module_annotation_group_line_numbers.append(line_number)
            if annotation["annotation_token"] == ".. toggle_name:":
                toggle_name = annotation["annotation_data"]
                self.current_module_annotated_toggle_names.add(toggle_name)
            elif annotation["annotation_token"] == ".. toggle_description:":
                toggle_description = annotation["annotation_data"].strip()
            elif annotation["annotation_token"] == ".. toggle_use_cases:":
                if "temporary" in annotation["annotation_data"]:
                    temporary_use_case = True
            elif annotation["annotation_token"] == ".. toggle_target_removal_date:":
                target_removal_date = annotation["annotation_data"]
            elif annotation["annotation_token"] == ".. toggle_default:":
                toggle_default = annotation["annotation_data"]

        if not toggle_name:
            self.add_message(
                self.NO_NAME_MESSAGE_ID,
                node=node,
                line=line_number,
            )
        if not toggle_description:
            self.add_message(
                self.EMPTY_DESCRIPTION_MESSAGE_ID,
                args=(toggle_name,),
                node=node,
                line=line_number,
            )
        if temporary_use_case and not target_removal_date:
            self.add_message(
                self.MISSING_TARGET_REMOVAL_DATE_MESSAGE_ID,
                args=(toggle_name,),
                node=node,
                line=line_number,
            )
        if toggle_default not in ["True", "False"]:
            self.add_message(
                self.NON_BOOLEAN_DEFAULT_VALUE,
                args=(toggle_name,),
                node=node,
                line=line_number,
            )

    @utils.only_required_for_messages(MISSING_ANNOTATION)
    def visit_call(self, node):
        """
        Check for missing annotations.
        """
        if self.is_annotation_missing(node):
            self.add_message(
                self.MISSING_ANNOTATION,
                node=node,
            )

    @utils.only_required_for_messages(INVALID_DJANGO_WAFFLE_IMPORT)
    def visit_import(self, node):
        if node.names[0][0] == "waffle":
            self.add_message(
                self.INVALID_DJANGO_WAFFLE_IMPORT,
                node=node,
            )

    @utils.only_required_for_messages(INVALID_DJANGO_WAFFLE_IMPORT)
    def visit_importfrom(self, node):
        if node.modname == "waffle":
            self.add_message(
                self.INVALID_DJANGO_WAFFLE_IMPORT,
                node=node,
            )

    def is_annotation_missing(self, node):
        """
        Check whether the node corresponds to a toggle instance creation. if yes, check that it is annotated.
        """
        if (
            not isinstance(node.func, Name)
            or node.func.name not in self.TOGGLE_FUNC_NAMES
        ):
            return False

        if not self.current_module_annotation_group_line_numbers:
            # There are no annotations left
            return True

        annotation_line_number = self.current_module_annotation_group_line_numbers[0]
        if annotation_line_number > node.tolineno:
            # The next annotation is located after the current node
            return True
        self.current_module_annotation_group_line_numbers.pop(0)

        # Check literal toggle name arguments
        if node.args and isinstance(node.args[0], Const) and isinstance(node.args[0].value, str):
            toggle_name = node.args[0].value
            if toggle_name not in self.current_module_annotated_toggle_names:
                return True
        return False


class SettingAnnotationChecker(AnnotationBaseChecker):
    """
    Perform checks on setting annotations.
    """

    CONFIG_FILENAMES = ["setting_annotations.yaml"]

    name = "setting-annotations"

    BOOLEAN_DEFAULT_VALUE = "setting-boolean-default-value"

    msgs = {
        ("E%d70" % BASE_ID): (
            "setting annotation (%s) cannot have a boolean value",
            BOOLEAN_DEFAULT_VALUE,
            "Setting with boolean values should be annotated as feature toggles",
        ),
    }

    @check_all_messages(msgs)
    def visit_module(self, node):
        """
        Run all checks on a single module.
        """
        self.check_module(node)

    def check_annotation_group(self, search, annotations, node):
        """
        Perform checks on a single annotation group.
        """
        if not annotations:
            return

        setting_name = ""
        setting_default = None
        line_number = None
        for annotation in annotations:
            if line_number is None:
                line_number = annotation["line_number"]
            if annotation["annotation_token"] == ".. setting_name:":
                setting_name = annotation["annotation_data"]
            elif annotation["annotation_token"] == ".. setting_default:":
                setting_default = annotation["annotation_data"]

        if setting_default in ["True", "False"]:
            self.add_message(
                self.BOOLEAN_DEFAULT_VALUE,
                args=(setting_name,),
                node=node,
                line=line_number,
            )



"""
Checks for unsafe ``yaml.load()`` calls.
The PyYAML library has not yet released a version
that addresses the vulnerability (listed below) in the commonly
called ``load()`` function.

Vulnerability: https://nvd.nist.gov/vuln/detail/CVE-2017-18342
PyYAML release plan: https://github.com/yaml/pyyaml/issues/193

When a new release is published, we may be able to remove this check.
"""

from pylint.checkers import BaseChecker, utils

from .common import BASE_ID, check_visitors


def register_checkers(linter):
    """Register checkers."""
    linter.register_checker(YamlLoadChecker(linter))


@check_visitors
class YamlLoadChecker(BaseChecker):
    """
    Checks for unsafe ``yaml.load()`` calls.
    """

    name = "yaml-load-checker"

    MESSAGE_ID = "unsafe-yaml-load"

    UNSAFE_CALLS = {"yaml.load", "yaml.load_all"}

    msgs = {
        f"C{BASE_ID}57": (
            "yaml.load%s() call is unsafe, use yaml.safe_load%s()",
            MESSAGE_ID,
            "yaml.load*() is unsafe",
        )
    }

    @utils.only_required_for_messages(MESSAGE_ID)
    def visit_call(self, node):
        """
        Check whether a call is an unsafe call to yaml.load.
        """
        func_name = node.func.as_string()
        if func_name in self.UNSAFE_CALLS:
            suffix = func_name.lstrip("yaml.load")
            self.add_message(self.MESSAGE_ID, args=(suffix, suffix), node=node)
